/*
 * Copyright (C) 2011 Technische Universitaet Muenchen
 * This file is part of the Sierpi project. For conditions of distribution and
 * use, please see the copyright notice in the file 'copyright.txt' at the root
 * directory of this package and the copyright notice at http://www5.in.tum.de/sierpi
 *
 *  Created on: Feb 14, 2011
 *      Author: Martin Schreiber (martin.schreiber@in.tum.de)
 */

#ifndef TRAVERSATORS_CEDGE_COMM_NORMALS_DEPTH_HPP_
#define TRAVERSATORS_CEDGE_COMM_NORMALS_DEPTH_HPP_


#include "libsierpi/stacks/CSimulationStacks.hpp"
#include "libsierpi/triangle/CTriangle_Factory.hpp"

namespace sierpi
{
namespace travs
{

/**
 * adaptive refinement
 */
template <class CKernelClass>
class CEdgeComm_Normals_Depth
{
public:
	/**
	 * implementation of kernel class
	 */
	CKernelClass kernelClass;

	/**
	 * a few typedefs which could be utilized by the traversator to get more informations
	 */
	typedef typename CKernelClass::TElementData TElementData;
	typedef typename CKernelClass::TEdgeData TEdgeData;
	typedef typename CKernelClass::TVertexScalar TVertexScalar;
	typedef CEdgeComm_Normals_Depth<CKernelClass> TThisClass;

	/**
	 * stack and lifo handlers used in this traversator
	 */
	CStack<char> *structure_stack_out;
	CStackReaderTopDown<char> structure_lifo_in;

	/**
	 * lifo and fifo iterator for element data
	 */
	CStackReaderTopDown<TElementData> element_data_lifo;
	CStackReaderBottomUp<TElementData> element_data_fifo;

	/**
	 * pointer to the corresponding communication stacks
	 */
	CStack<TEdgeData> *edge_data_comm_left_edge_stack;
	CStack<TEdgeData> *edge_data_comm_right_edge_stack;

	/**
	 * pointer to stack used for edge communication in forward order
	 */
	CStack<TEdgeData> *edge_comm_buffer;


#define SIERPI_PARALLEL_CODE	1
#include "auto/CTraversator_EdgeComm_Normals_Depth_autogenerated.hpp"
#undef SIERPI_PARALLEL_CODE
#include "../CTraversatorClassInc_SetupTwoPass.hpp"


	void action_FirstPass(
			CSimulationStacks<CTsunamiElementData, CTsunamiEdgeData> *p_cSimulationStacks
	)
	{
		assert(p_cSimulationStacks->structure_stacks.direction == p_cSimulationStacks->element_data_stacks.direction);
		assert(p_cSimulationStacks->structure_stacks.direction == CFBStacks<char>::FORWARD);

		structure_stack_out = &p_cSimulationStacks->structure_stacks.backward;
		assert(structure_stack_out->isEmpty());

		structure_lifo_in.setup(p_cSimulationStacks->structure_stacks.forward);

		assert(!p_cSimulationStacks->structure_stacks.forward.isEmpty());

		edge_data_comm_left_edge_stack = &p_cSimulationStacks->edge_data_comm_left_edge_stack;
		edge_data_comm_right_edge_stack = &p_cSimulationStacks->edge_data_comm_right_edge_stack;

		edge_comm_buffer = &p_cSimulationStacks->edge_comm_buffer;
		element_data_lifo.setup(p_cSimulationStacks->element_data_stacks.forward);

		kernelClass.traversal_pre_hook();

		(this->*sfcRecursiveMethod_Forward)(triangleFactory.recursionDepthFirstRecMethod);

		kernelClass.traversal_post_hook();

		assert(structure_lifo_in.isEmpty());
		assert(element_data_lifo.isEmpty());
	}



	/**
	 * second pass for parallel processing (exchange stacks)
	 */
	void action_SecondPass_Parallel(
			CSimulationStacks<CTsunamiElementData, CTsunamiEdgeData> *p_cSimulationStacks
	)
	{
		structure_lifo_in.setup(p_cSimulationStacks->structure_stacks.backward);
		element_data_fifo.setup(p_cSimulationStacks->element_data_stacks.forward);

		assert(!p_cSimulationStacks->structure_stacks.backward.isEmpty());


		// clear "old" edge comm stacks
		p_cSimulationStacks->edge_data_comm_left_edge_stack.clear();
		p_cSimulationStacks->edge_data_comm_right_edge_stack.clear();

		edge_data_comm_left_edge_stack = &p_cSimulationStacks->edge_data_comm_exchange_right_edge_stack;
		edge_data_comm_right_edge_stack = &p_cSimulationStacks->edge_data_comm_exchange_left_edge_stack;

		edge_comm_buffer = &p_cSimulationStacks->edge_comm_buffer;
		element_data_lifo.setup(p_cSimulationStacks->element_data_stacks.forward);

		kernelClass.traversal_pre_hook();

		(this->*sfcRecursiveMethod_Backward)(triangleFactory.recursionDepthFirstRecMethod);

		kernelClass.traversal_post_hook();

		assert(edge_comm_buffer->isEmpty());
		assert(edge_data_comm_left_edge_stack->isEmpty());
		assert(edge_data_comm_right_edge_stack->isEmpty());
		assert(structure_lifo_in.isEmpty());
		assert(element_data_fifo.isEmpty());

		structure_stack_out->clear();

	}


	/**
	 * second pass for serial traversal
	 */
	void action_SecondPass_Serial(
			CSimulationStacks<CTsunamiElementData, CTsunamiEdgeData> *p_cSimulationStacks
	)
	{
		structure_lifo_in.setup(p_cSimulationStacks->structure_stacks.backward);
		element_data_fifo.setup(p_cSimulationStacks->element_data_stacks.forward);

		assert(!p_cSimulationStacks->structure_stacks.backward.isEmpty());

		edge_data_comm_left_edge_stack = &p_cSimulationStacks->edge_data_comm_right_edge_stack;
		edge_data_comm_right_edge_stack = &p_cSimulationStacks->edge_data_comm_left_edge_stack;

		edge_comm_buffer = &p_cSimulationStacks->edge_comm_buffer;
		element_data_lifo.setup(p_cSimulationStacks->element_data_stacks.forward);

		kernelClass.traversal_pre_hook();
		(this->*sfcRecursiveMethod_Backward)(triangleFactory.recursionDepthFirstRecMethod);

		kernelClass.traversal_post_hook();

		assert(edge_comm_buffer->isEmpty());
		assert(edge_data_comm_left_edge_stack->isEmpty());
		assert(edge_data_comm_right_edge_stack->isEmpty());
		assert(structure_lifo_in.isEmpty());
		assert(element_data_fifo.isEmpty());

		structure_stack_out->clear();

	}


};

}
}

#endif /* CSTRUCTURE_ADAPTIVEREFINEMENT_HPP_ */
