/*
 * Copyright (C) 2011 Technische Universitaet Muenchen
 * This file is part of the Sierpi project. For conditions of distribution and
 * use, please see the copyright notice in the file 'copyright.txt' at the root
 * directory of this package and the copyright notice at http://www5.in.tum.de/sierpi
 *
 *  Created on: Dec 15, 2010
 *      Author: Martin Schreiber (martin.schreiber@in.tum.de)
 */

#define COMPILE_SIMULATION_WITH_GUI	0

#include <stdlib.h>
#include <stdint.h>
#include <sstream>

#if COMPILE_WITH_OMP
	#include <omp.h>
#endif

#if COMPILE_WITH_TBB
#include <tbb/task_scheduler_init.h>
#endif

#include "config.h"
#include "lib/iRef.hpp"
#include "libsierpi/stacks/CSimulationStacks.hpp"
#include "lib/CProcessMemoryInformation.hpp"
#include "lib/CStopwatch.hpp"

#if SIMULATION_TSUNAMI_SERIAL
	#include "simulations/tsunami_serial/CSimulationTsunami_Serial.hpp"
	#define SIMULATION_CLASS	CSimulationTsunami
#endif

#if SIMULATION_TSUNAMI_SERIAL_REGULAR
	#include "simulations/tsunami_serial_regular/CSimulationTsunami_Serial_Regular.hpp"
	#define SIMULATION_CLASS	CSimulationTsunami
#endif

#if SIMULATION_TSUNAMI_PARALLEL
	#include "simulations/tsunami_parallel/CSimulationTsunami_Parallel.hpp"
	#define SIMULATION_CLASS	CSimulationTsunami_Parallel
#endif

#if COMPILE_WITH_IOMP
	#include <OR.hpp>
	#include <AND.hpp>
	#include <Claim.hpp>
	#include <PEQuantity.hpp>
#endif


CProcessMemoryInformation cProcessMemoryInformation;


int timesteps = 13000;
float timestepSize = -1;

int initialRecursionDepth = 6;
int minRelativeRecursionDepth = 0;
int maxRelativeRecursionDepth = 6;

int number_of_threads_to_use = -1;

bool divide_number_of_timesteps_by_timestep_size = false;

//uint64_t partition_split_workload_size = 16;
uint64_t partition_split_workload_size = 512;
uint64_t partition_join_workload_size = 0;

int worldSceneId = 3;

int verbose = 0;

int output_simulation_data_each_nth_timestep = -1;
const char *output_simulation_data_filename = "frame_%08i.vtk";
bool output_partitions_to_vtk_file = false;


#if COMPILE_WITH_TBB
    tbb::task_scheduler_init *tbb_task_scheduler_init;
#endif


#if 0
extern long int random (void)	throw()
{
	std::cout << "RANDOM FUNCTION CALLED!" << std::endl;
	exit(-1);
}
#endif



void run()
{
	SIMULATION_CLASS cSimulation;

	if (verbose > 2)
	{
		cSimulation.outputVerboseInformation();
	}

	/**************************
	 * SETUP
	 */
#if SIMULATION_TSUNAMI_PARALLEL
	cSimulation.setWorldSceneId(worldSceneId);
#endif

	cSimulation.setInitialRecursionDepth(initialRecursionDepth);
	cSimulation.setMinRelativeRecursionDepth(minRelativeRecursionDepth);
	cSimulation.setMaxRelativeRecursionDepth(maxRelativeRecursionDepth);

	cSimulation.setSplitJoinSizes(partition_split_workload_size, partition_join_workload_size);

	cSimulation.reset_Simulation();

	cSimulation.setTimestepSize(timestepSize);

	if (divide_number_of_timesteps_by_timestep_size)
	{
		if (verbose > 2)
			std::cout << "Updating number of timesteps from " << timesteps << " to ";

		timesteps = (int)((double)timesteps/(double)cSimulation.timestep_size);

		if (verbose > 2)
			std::cout << timesteps << std::endl;

		if (verbose > 2)
			std::cout << "Timesteps: " << timesteps << std::endl;

		if (verbose > 2)
			std::cout << "Updating number of output of simulation data from " << output_simulation_data_each_nth_timestep << " to ";
			output_simulation_data_each_nth_timestep = (int)((double)output_simulation_data_each_nth_timestep/(double)cSimulation.timestep_size);

		if (verbose > 2)
			std::cout << "Output_simulation_data_each_nth_timestep: " << output_simulation_data_each_nth_timestep;

		std::cout << std::endl;
	}


	if (verbose > 2)
	{
		std::cout << " + Number of threads to use: " << number_of_threads_to_use << std::endl;
		std::cout << " + Initial recursion depth: " << initialRecursionDepth << std::endl;
		std::cout << " + Verbose: " << verbose << std::endl;
		std::cout << " + Timesteps: " << timesteps << std::endl;
		std::cout << " + Timestep size: " << cSimulation.timestep_size << std::endl;
		std::cout << " + Output .vtk files each #nth timestep: " << output_simulation_data_each_nth_timestep << std::endl;
		std::cout << " + Output simulation data filename: " << output_simulation_data_filename << std::endl;
		std::cout << " + InitialDepth: " << initialRecursionDepth << std::endl;
		std::cout << " + MinDepth: " << (initialRecursionDepth+minRelativeRecursionDepth) << std::endl;
		std::cout << " + MaxDepth: " << (initialRecursionDepth+maxRelativeRecursionDepth) << std::endl;
		std::cout << " + WorldSceneId: " << worldSceneId << std::endl;
		std::cout << " + Partition size when to request split: " << partition_split_workload_size << std::endl;
		std::cout << " + Partition size when to request join (both childs have to request a join): " << partition_join_workload_size << std::endl;

#if ADAPTIVE_SUBPARTITION_STACKS
		std::cout << " + ADAPTIVE_SUBPARTITION_STACKS enabled" << std::endl;
#endif
	}

	/**************************
	 * setup column
	 */
	if (verbose > 0)
		std::cout << "[ SETUP ]" << std::endl;

	if (verbose > 1)
		std::cout << " + initial number of triangles: " << cSimulation.number_of_triangles << std::endl;


	cSimulation.number_of_triangles = cSimulation.setup_ColumnAt2DPosition(-0.5, 0.4, 0.2);

	if (verbose > 1)
		std::cout << " + refined to " << cSimulation.number_of_triangles << " triangles" << std::endl;

	if (verbose > 0)
		std::cout << " + pre splitting partitions" << std::endl;


	/****************************
	 * setup initial split of partitions
	 */
	{
		size_t t = 0;
		while (cSimulation.number_of_simulation_handler_partitions != t)
		{
			t = cSimulation.number_of_simulation_handler_partitions;

#if SIMULATION_TSUNAMI_PARALLEL
			cSimulation.setup_SplitJoinPartitions();
#endif

			if (verbose > 1)
				std::cout << " + splitted to " << cSimulation.number_of_simulation_handler_partitions << " partitions" << std::endl;
		}
	}

	CStopwatch cStopwatch;


	if (verbose > 1)
	{
		std::cout << "[ START ]" << std::endl;
	}

	cStopwatch.start();

	/**************************
	 * SIMULATION
	 */
	uint64_t sum_number_of_triangles = 0;
	uint64_t sum_number_of_partitions = 0;
	char char_buffer[1024];

	int outputFrame = 0;
	int frameCounter = 0;

	double edgeCommTime = 0;
	double adaptiveTime = 0;
	double splitJoinTime = 0;

	/**
	 * Simulation loop
	 */
	if (number_of_threads_to_use == -1)
	{
#if COMPILE_WITH_OMP

#pragma omp parallel
#pragma omp single
		number_of_threads_to_use = omp_get_max_threads();

#elif COMPILE_WITH_TBB
		number_of_threads_to_use = 1;
#else
		number_of_threads_to_use = 1;
#endif
	}

#if COMPILE_WITH_IOMP
	int max_num_threads = omp_get_max_threads();
	size_t initial_number_of_triangles = cSimulation.number_of_triangles;

	Claim claim;
#endif

	for (int i = 0; i < timesteps; i++)
	{
#if COMPILE_WITH_IOMP
#if 0
		/*
		 * increase number of threads to maximum when exceeding twice the
		 * initial number of triangles
		 */
		int num_threads;
//		std::cout << initial_number_of_triangles*2 << " > " << cSimulation.number_of_triangles << std::endl;

		if (initial_number_of_triangles*2 > cSimulation.number_of_triangles)
			num_threads = 1;
		else
			num_threads = max_num_threads;
#else
		int num_threads = max_num_threads;
#endif

//		std::cout << num_threads << " / " << max_num_threads << std::endl;

		claim.invade(PEQuantity(1, num_threads));

	#if COMPILE_WITH_OMP
	#pragma omp parallel
	#pragma omp single nowait
	#endif
#endif
		{
			/**
			 * output some data?
			 */
			if (output_simulation_data_each_nth_timestep > 0)
			{
				if (outputFrame == 0)
				{
					sprintf(char_buffer, output_simulation_data_filename, frameCounter);

					if (verbose > 3)
					{
						std::cout << "=========================================" << std::endl;
						if (i != 0)
							std::cout << "   + " << ((double)sum_number_of_triangles/cStopwatch.getTimeSinceStart())*0.000001 << " MTPS(Mega Triangles per second)" << std::endl;
						std::cout << "   + writing file " << char_buffer << std::endl;
						std::cout << "   + " << i << "\tTIMESTEP" << std::endl;
						std::cout << "   + " << cSimulation.number_of_triangles << "\tTRIANGLES" << std::endl;
						double elementdata_megabyte_per_timestep = ((double)sizeof(CTsunamiElementData)*2.0)*((double)cSimulation.number_of_triangles)/(1024.0*1024.0);
						std::cout << "   + " << elementdata_megabyte_per_timestep << "\tElementData Megabyte per Timestep (RW)" << std::endl;
						std::cout << "   + " << cSimulation.number_of_simulation_handler_partitions << "\tPARTITIONS" << std::endl;
					}

					cSimulation.writeTrianglesToVTKFile(char_buffer);
					std::string partitionFile = "partitions_";
					partitionFile += char_buffer;
#if SIMULATION_TSUNAMI_PARALLEL
					if (output_partitions_to_vtk_file)
						cSimulation.writePartitionsToVTKFile(partitionFile.c_str());
#endif
					frameCounter++;
				}

				outputFrame = (outputFrame+1) % output_simulation_data_each_nth_timestep;
			}
			else if (verbose > 3)
			{
				if (outputFrame == 0)
				{
					std::cout << "=========================================" << std::endl;
					if (i != 0)
						std::cout << "   + " << ((double)sum_number_of_triangles/cStopwatch.getTimeSinceStart())*0.000001 << " MTPS(Mega Triangles per second)" << std::endl;
					std::cout << "   + " << i << "\tTIMESTEP" << std::endl;
					std::cout << "   + " << cSimulation.number_of_triangles << "\tTRIANGLES" << std::endl;
					double elementdata_megabyte_per_timestep = ((double)sizeof(CTsunamiElementData)*2.0)*((double)cSimulation.number_of_triangles)/(1024.0*1024.0);
					std::cout << "   + " << elementdata_megabyte_per_timestep << "\tElementData Megabyte per Timestep (RW)" << std::endl;
					std::cout << "   + " << cSimulation.number_of_simulation_handler_partitions << "\tPARTITIONS" << std::endl;

					if (verbose > 5)
					{
						cProcessMemoryInformation.outputUsageInformation();
					}
				}

				outputFrame = (outputFrame+1) % 100;
			}


			/*
			 * single simulation timestep
			 */
			if (verbose > 2)
			{
				cSimulation.runSingleTimestepDetailedBenchmarks(
						&edgeCommTime,
						&adaptiveTime,
						&splitJoinTime
					);
			}
			else
			{
				cSimulation.runSingleTimestep();
			}

			/*
			 * increment counter of number of triangles processed so far
			 */
			sum_number_of_triangles += cSimulation.number_of_triangles;
			sum_number_of_partitions += cSimulation.number_of_simulation_handler_partitions;

			/*
			 * verbose points
			 */
			if (verbose > 5)
			{
				std::cout << "." << std::flush;
			}
		}	// end parallel region iOMP
	}

	cStopwatch.stop();

	if (verbose > 1)
	{
		std::cout << "[ END ]" << std::endl;
		std::cout << std::endl;
	}


	if (verbose > 2)
	{
		std::cout << std::endl;
		std::cout << "Timings for simulation phases:" << std::endl;
		std::cout << " + EdgeCommTime: " << edgeCommTime << std::endl;
		std::cout << " + AdaptiveTime: " << adaptiveTime << std::endl;
		std::cout << " + SplitJoinTime: " << splitJoinTime << std::endl;
		std::cout << std::endl;
	}


	double stoptime = cStopwatch();

	std::cout << timesteps << " TS (Timesteps)" << std::endl;
	std::cout << cSimulation.timestep_size << " TSS (Timestep size)" << std::endl;
	std::cout << stoptime << " SFS (Overall seconds for Simulation)" << std::endl;
	std::cout << sum_number_of_triangles << " TP (Triangles processed)" << std::endl;
	std::cout << (double)stoptime/(double)timesteps << " ASPT (Averaged Seconds per Timestep)" << std::endl;
	std::cout << (double)sum_number_of_triangles/(double)timesteps << " TPST (Triangles Processed in Average per Simulation Timestep)" << std::endl;

#if SIMULATION_TSUNAMI_PARALLEL
	std::cout << (double)sum_number_of_partitions/(double)timesteps << " PPST (Partitions Processed in Average per Simulation Timestep)" << std::endl;
#endif

	std::cout << ((double)sum_number_of_triangles/cStopwatch())*0.000001 << " MTPS (Million Triangles per Second)" << std::endl;
	std::cout << ((double)sum_number_of_triangles/(cStopwatch()*(double)number_of_threads_to_use))*0.000001 << " MTPSPT (Million Triangles per Second per Thread)" << std::endl;

	double elementdata_megabyte_per_timestep = ((double)sizeof(CTsunamiElementData)*2.0)*((double)sum_number_of_triangles/(double)timesteps)/(1024.0*1024.0);
	std::cout << elementdata_megabyte_per_timestep << " EDMBPT (ElementData Megabyte per Timestep (RW))" << std::endl;

	double elementdata_megabyte_per_second = ((double)sizeof(CTsunamiElementData)*2.0)*((double)sum_number_of_triangles/(double)stoptime)/(1024.0*1024.0);
	std::cout << elementdata_megabyte_per_second << " EDMBPS (ElementData Megabyte per Second (RW))" << std::endl;
}



int main(int argc, char *argv[])
{
	char optchar;
	while ((optchar = getopt(argc, argv, "d:t:f:w:a:i:s:v:n:o:l:rpg:")) > 0)
	{
		switch(optchar)
		{
		case 'r':
			divide_number_of_timesteps_by_timestep_size = true;
			break;

		case 'p':
			output_partitions_to_vtk_file = true;
			break;

		case 'd':
			initialRecursionDepth = atoi(optarg);
			break;

		case 'v':
			verbose = atoi(optarg);
			break;

		case 't':
			timesteps = atoi(optarg);
			break;

		case 's':
			timestepSize = atof(optarg);
			break;

		case 'f':
			output_simulation_data_each_nth_timestep = atoi(optarg);
			break;

		case 'g':
			output_simulation_data_filename = optarg;
			break;

		case 'n':
			number_of_threads_to_use = atoi(optarg);
			break;

		case 'w':
			worldSceneId = atoi(optarg);
			break;

		case 'i':
			minRelativeRecursionDepth = atoi(optarg);
			break;

		case 'a':
			maxRelativeRecursionDepth = atoi(optarg);
			break;

		case 'o':
			partition_split_workload_size = atoi(optarg);
			break;

		case 'l':
			partition_join_workload_size = atoi(optarg);
			break;

		default:
			goto parameter_error;
		}

	}

	goto parameter_error_ok;

	parameter_error:
	std::cout << "usage: " << argv[0] << std::endl;
	std::cout << "	[-d [int]: initial recursion depth]" << std::endl;
	std::cout << "	[-v [int]: verbose mode (0-5)]" << std::endl;
	std::cout << "	[-n [int]: number of threads to use]" << std::endl;
	std::cout << "	[-t [int]: timesteps]" << std::endl;
	std::cout << "	[-r : divide number of timesteps and frame output rate by timestep size]" << std::endl;
	std::cout << "	[-s [float]: timestep size]" << std::endl;
	std::cout << "	[-p : also write vtk files with partitions]" << std::endl;
	std::cout << "	[-f [int]: output .vtk files each #nth timestep]" << std::endl;
	std::cout << "	[-g [string]: outputfilename, default: frame_%08i.vtk" << std::endl;
	std::cout << "	[-w [int]: world scene]" << std::endl;
	std::cout << "	[-i [int]: min relative recursion depth]" << std::endl;
	std::cout << "	[-a [int]: max relative recursion depth]" << std::endl;
	std::cout << "	[-o [int]: partition size when to request split]" << std::endl;
	std::cout << "	[-l [int]: partition size when to request join (both childs have to request a join)]" << std::endl;

	return -1;

parameter_error_ok:
	if (verbose > 5)
	{
		cProcessMemoryInformation.outputUsageInformation();
	}
	srand(0);


#if COMPILE_WITH_TBB
	// setup scheduler
	tbb_task_scheduler_init = new tbb::task_scheduler_init(number_of_threads_to_use);

	if (verbose > 5)
		std::cout << "tbb::task_scheduler_init::default_num_threads(): " << tbb::task_scheduler_init::default_num_threads() << std::endl;
#endif


#if COMPILE_WITH_OMP
	if (number_of_threads_to_use == -1)
		number_of_threads_to_use = omp_get_num_threads();
	else
		omp_set_num_threads(number_of_threads_to_use);

	if (verbose > 5)
		std::cout << "omp_get_max_threads(): " << omp_get_max_threads() << std::endl;
#endif

	if (partition_split_workload_size > 0 && partition_split_workload_size < 2)
	{
		std::cout << "Invalid split size: " << partition_split_workload_size << "   setting split to 2" << std::endl;
		partition_split_workload_size = 2;
	}

	if (partition_join_workload_size == 0)
	{
		partition_join_workload_size = partition_split_workload_size/2;
	}
	else
	{
		if (partition_join_workload_size > partition_split_workload_size)
		{
			std::cerr << "ERROR: partition_join_workload_size > partition_split_workload_size" << std::endl;
			return -1;
		}
	}

	/*
	 * RUN THE SIMULATION WITH PARALLELIZATION
	 */
#if COMPILE_WITH_IOMP
	/*
	 * we have to move the parallel region into the simulation
	 * timestep loop to control the amount of CPUs which are used
	 */
#else
	#if COMPILE_WITH_OMP
	#pragma omp parallel
	#pragma omp single nowait
	#endif
#endif
	{
		run();
	}

#if COMPILE_WITH_TBB
		delete tbb_task_scheduler_init;
#endif


#ifdef DEBUG
	if (!debug_ibase_list.empty())
	{
		std::cout << "MEMORY LEAK: iBase class missing in action" << std::endl;
		std::cout << "  + number of classes: " << debug_ibase_list.size() << std::endl;
	}
#endif

	if (verbose > 5)
	{
		cProcessMemoryInformation.outputUsageInformation();
	}

	return 0;
}
