/*
 * Copyright (C) 2011 Technische Universitaet Muenchen
 * This file is part of the Sierpi project. For conditions of distribution and
 * use, please see the copyright notice in the file 'copyright.txt' at the root
 * directory of this package and the copyright notice at http://www5.in.tum.de/sierpi
 *
 * CFluxLaxFriedrich.hpp
 *
 *  Created on: Dec 20, 2011
 *      Author: schreibm
 */

#ifndef CFLUXSOLVER_LAXFRIEDRICH_HPP_
#define CFLUXSOLVER_LAXFRIEDRICH_HPP_


template <typename T>
class CFluxSolver_LaxFriedrich
{
public:
	/**
	 * \brief compute the flux and store the result to o_flux
	 *
	 * this method uses a lax friedrichs flux
	 *
	 * the input data is assumed to be rotated to the edge normal pointing along the positive part of the x axis
	 */
	void computeNetUpdates(
			CTsunamiEdgeData &i_edgeData_left,		///< edge data on left (left) edge
			CTsunamiEdgeData &i_edgeData_right,		///< edge data on right (outer) edge
			CTsunamiEdgeData &o_edgeFlux_left,		///< output for left flux
			CTsunamiEdgeData &o_edgeFlux_right,		///< output for outer flux
			T &o_max_wave_speed,					///< maximum wave speed
			T i_gravitational_constant				///< gravitational constant
	)
	{
		T left_vx = i_edgeData_left.qx/i_edgeData_left.h;
		T left_vy = i_edgeData_left.qy/i_edgeData_left.h;

		T right_vx = i_edgeData_right.qx/i_edgeData_right.h;
		T right_vy = i_edgeData_right.qy/i_edgeData_right.h;

#if 0
		/**
		 * TODO: update to non-intrinsic version below
		 */
// 4 ops
		__m128 m_left_right_qx_qy = _mm_set_ps(i_edgeData_right.qy, i_edgeData_right.qx, i_edgeData_left.qy, i_edgeData_left.qx);
		__m128 m_left_h_right_h = _mm_set_ps(i_edgeData_right.h, i_edgeData_right.h, i_edgeData_left.h, i_edgeData_left.h);
		__m128 m_left_right_vx_vy = _mm_div_ps(m_left_right_qx_qy, m_left_h_right_h);

		__m128 m_gravity = _mm_set1_ps(i_gravitational_constant);

		/*
		T left_lambda = CMath::sqrt<T>(left_vx*left_vx+left_vy*left_vy) + CMath::sqrt<T>(config.gravity*i_edgeData_left.h);
		T right_lambda = CMath::sqrt<T>(right_vx*right_vx+right_vy*right_vy) + CMath::sqrt<T>(config.gravity*i_edgeData_right.h);
		*/

// 7*2 ops (scalar)
		__m128 m_left_up = _mm_dp_ps(m_left_right_vx_vy, m_left_right_vx_vy, 16*3 + 7);
		__m128 m_left_bottom = _mm_dp_ps(m_left_right_vx_vy, m_left_right_vx_vy, 16*12 + 7);

		__m128 m_right_side = _mm_mul_ps(m_gravity, m_left_h_right_h);

		__m128 m_sqrt_helper1 = _mm_blend_ps(m_left_up, m_right_side, 2);		// r0, r1
		__m128 m_sqrt_helper2 = _mm_blend_ps(m_left_bottom, m_right_side, 8);	// r2, r3
		__m128 m_sqrt = _mm_sqrt_ps(_mm_blend_ps(m_sqrt_helper1, m_sqrt_helper2, 12));

		T sqrt_stuff[4];
		_mm_store_ps(sqrt_stuff, m_sqrt);

		/*
		T lambda = CMath::max(left_lambda, right_lambda);
		*/
		T lambda = CMath::max(sqrt_stuff[0] + sqrt_stuff[1], sqrt_stuff[2] + sqrt_stuff[3]);

#else

#if 1
		/**
		 * compute eigenvalues for left and outer DOFs
		 */
		T left_grav_h = CMath::sqrt(i_gravitational_constant*i_edgeData_left.h);
		T right_grav_h = CMath::sqrt(i_gravitational_constant*i_edgeData_right.h);

		T left_lambda_add = CMath::abs(left_vx + left_grav_h);
		T left_lambda_sub = CMath::abs(left_vx - left_grav_h);

		T right_lambda_add = CMath::abs(right_vx + right_grav_h);
		T right_lambda_sub = CMath::abs(right_vx - right_grav_h);

		T left_lambda = CMath::max(left_lambda_add, left_lambda_sub);
		T right_lambda = CMath::max(right_lambda_add, right_lambda_sub);

#else

		T left_lambda = CMath::sqrt(left_vx*left_vx+left_vy*left_vy) + CMath::sqrt(config.gravity*i_left.h);
		T right_lambda = CMath::sqrt(right_vx*right_vx+right_vy*right_vy) + CMath::sqrt(config.gravity*i_outer.h);

#endif

		T lambda = CMath::max(left_lambda, right_lambda);

#endif

		o_edgeFlux_left.h =
				(T)0.5*
				(
					i_edgeData_left.qx + i_edgeData_right.qx
					+ (T)lambda*(i_edgeData_left.h - i_edgeData_right.h)
				);

		o_edgeFlux_left.qx =
				(T)0.5*
				(
					left_vx*i_edgeData_left.qx +
					right_vx*i_edgeData_right.qx +
					(TTsunamiDataScalar)0.5*i_gravitational_constant*(i_edgeData_left.h*i_edgeData_left.h + i_edgeData_right.h*i_edgeData_right.h)
					+
					(T)lambda*(
							i_edgeData_left.qx - i_edgeData_right.qx
					)
				);

		o_edgeFlux_left.qy =
				(T)0.5*
				(
					left_vx*i_edgeData_left.qx +
					right_vx*i_edgeData_right.qx +
					(
							left_vy*i_edgeData_left.qx +
							right_vy*i_edgeData_right.qx
					) +
					0*(T)lambda*(
							i_edgeData_left.qy - i_edgeData_right.qy
					)
				);

		o_edgeFlux_right.h = -o_edgeFlux_left.h;
		o_edgeFlux_right.qx = -o_edgeFlux_left.qx;
		o_edgeFlux_right.qy = -o_edgeFlux_left.qy;
	}
};



#endif /* CFLUXLAXFRIEDRICH_HPP_ */
