/*
 * Copyright (C) 2011 Technische Universitaet Muenchen
 * This file is part of the Sierpi project. For conditions of distribution and
 * use, please see the copyright notice in the file 'copyright.txt' at the root
 * directory of this package and the copyright notice at http://www5.in.tum.de/sierpi
 *
 *  Created on: Feb 14, 2011
 *      Author: Martin Schreiber (martin.schreiber@in.tum.de)
 */

#ifndef CEDGECOMM_TSUNAMI_0TH_ORDER_HPP_
#define CEDGECOMM_TSUNAMI_0TH_ORDER_HPP_


#include "config.h"

#include "libmath/CMath.hpp"
#include "libmath/CVertex2d.hpp"
#include "simulations/tsunami_common/types/CTsunamiTypes.hpp"

#include "EEdgeComm_Tsunami_BoundaryConditions.hpp"

#include "libsierpi/traversators/edgeComm/CEdgeComm_Normals_Depth.hpp"

#include "../../flux_solver/CFluxSolver.hpp"
#include "CFluxProjections.hpp"

namespace sierpi
{
namespace kernels
{

class CEdgeComm_Tsunami_0thOrder_config
{
public:
	TTsunamiDataScalar delta_timestep;
	TTsunamiDataScalar square_side_length;
	TTsunamiDataScalar gravitational_constant;

	CTsunamiEdgeData boundary_dirichlet;
};



/**
 * \brief class implementation for computation of 1st order DG shallow water simulation
 *
 * for DG methods, we basically need 3 different methods:
 *
 * 1) Storing the edge communication data (storeLeftEdgeCommData, storeRightEdgeCommData,
 *    storeHypEdgeCommData) which has to be sent to the adjacent triangles
 *
 * 2) Computation of the numerical fluxes (computeFlux) based on the local edge comm data
 *    and the transmitted DOF's from the previous operation.
 *
 *    TODO: since those fluxes are symmetrical, only one flux computation is needed for
 *          one edge.
 *
 * 3) Computation of the element update based on the fluxes computed in the previous step
 *    (elementAction)
 */
class CEdgeComm_Tsunami_0thOrder	:
	public CFluxProjections<TTsunamiDataScalar>
{
public:
	typedef TTsunamiDataScalar TVertexScalar;
	typedef CVertex2d<TTsunamiDataScalar> TVertexType;
	typedef CTsunamiElementData TElementData;
	typedef CTsunamiEdgeData TEdgeData;
	typedef TTsunamiDataScalar T;

	// TRAVersator typedef
	typedef sierpi::travs::CEdgeComm_Normals_Depth<CEdgeComm_Tsunami_0thOrder> TRAV;

	/**
	 * generic configurations (gravitational constant, timestep size, ...)
	 */
	CEdgeComm_Tsunami_0thOrder_config config;

	/**
	 * true if an instability was detected
	 */
	bool instabilityDetected;

	/**
	 * accessor to flux solver
	 */
	CFluxSolver<T> fluxSolver;

	/**
	 * available boundary conditions
	 */
	enum EBoundaryCondition
	{
		BOUNDARY_CONDITION_DIRICHLET,
		BOUNDARY_CONDITION_VELOCITY_DAMPING,
		BOUNDARY_CONDITION_BOUNCE_BACK
	};

	/**
	 * set-up boundary condition
	 */
	EBoundaryCondition eBoundaryCondition;

	/**
	 * damping factor when velocity damping boundary condition is used
	 */
	TTsunamiDataScalar eBoundaryConditionVelocityDampingFactor;

	/**
	 * return the set-up timestep size
	 */
	TTsunamiDataScalar getTimestepSize()
	{
		return config.delta_timestep;
	}



	CEdgeComm_Tsunami_0thOrder()	:
		eBoundaryCondition(BOUNDARY_CONDITION_DIRICHLET),
//		eBoundaryCondition(BOUNDARY_CONDITION_VELOCITY_DAMPING),
//		eBoundaryCondition(BOUNDARY_CONDITION_BOUNCE_BACK),
		eBoundaryConditionVelocityDampingFactor(0.9)
	{
		config.boundary_dirichlet.h = CMath::numeric_inf<T>();
		config.boundary_dirichlet.qx = CMath::numeric_inf<T>();
		config.boundary_dirichlet.qy = CMath::numeric_inf<T>();
	}



	/**
	 * \brief setup parameters for simulation
	 */
	void setup_WithParameters(
			T i_timestep_size,					///< timestep size
			T i_square_side_length,				///< length of a square (a single catheti)
			T i_gravitational_constant = 9.81	///< gravitational constant
	)
	{
		config.delta_timestep = i_timestep_size;
		config.gravitational_constant = i_gravitational_constant;
		config.square_side_length = i_square_side_length;
	}


	/**
	 * \brief setup parameters using child's/parent's kernel
	 */
	void setup_WithKernel(
			CEdgeComm_Tsunami_0thOrder &parent
	)
	{
		config = parent.config;
		eBoundaryCondition = parent.eBoundaryCondition;
		eBoundaryConditionVelocityDampingFactor = parent.eBoundaryConditionVelocityDampingFactor;
	}


	/**
	 * setup parameter for dirichlet boundary condition
	 */
	void setup_WithParameters_KernelBoundaryDirichlet(
			const CTsunamiEdgeData *p_boundary_dirichlet
	)
	{
		config.boundary_dirichlet = *p_boundary_dirichlet;
	}



	/**
	 * \brief CALLED by TRAVERSATOR: this method is executed before traversal of the sub-partition
	 */
	void traversal_pre_hook()
	{
		/**
		 * set instability detected to false during every new traversal.
		 *
		 * this variable is used to avoid excessive output of instability
		 * information when an instability is detected.
		 */
		instabilityDetected = false;
	}


	/**
	 * \brief CALLED by TRAVERSATOR: this method is executed after traversal of the sub-partition
	 */
	void traversal_post_hook()
	{
	}



	/**
	 * \brief CALLED by TRAVERSATOR: Store data send via hypotenuse edge
	 *
	 * this method is responsible for storing the edge communication data
	 * across the hypotenuse.
	 *
	 * depending on the implemented elementData operations, the edge communication
	 * data has to be rotated to the normal space.
	 */
	inline void storeHypCommData(
			TVertexScalar i_hyp_normal_x, TVertexScalar i_hyp_normal_y,
			TVertexScalar i_right_normal_x, TVertexScalar i_right_normal_y,
			TVertexScalar i_left_normal_x, TVertexScalar i_left_normal_y,
			int i_depth,

			const CTsunamiElementData *i_elementData,
			CTsunamiEdgeData *o_edge_comm_data
	)
	{
		*o_edge_comm_data = i_elementData->dofs;
		projectToNormalSpaceAndInvert(o_edge_comm_data->qx, o_edge_comm_data->qy, i_hyp_normal_x, i_hyp_normal_y);

#if COMPILE_WITH_DEBUG_TSUNAMI_ELEMENTDATA_VALIDATION
		i_elementData->validation.setupHypEdgeCommData(&(o_edge_comm_data->validation));
		o_edge_comm_data->validation.setupNormal(i_hyp_normal_x, i_hyp_normal_y);
#endif
	}


	/**
	 * \brief CALLED by TRAVERSATOR: Store data send via right edge
	 *
	 * this method is responsible for storing the edge communication data
	 * across the right edge.
	 *
	 * depending on the implemented elementData operations, the edge communication
	 * data has to be rotated to the normal space.
	 */
	inline void storeRightEdgeCommData(
			TVertexScalar i_hyp_normal_x, TVertexScalar i_hyp_normal_y,
			TVertexScalar i_right_normal_x, TVertexScalar i_right_normal_y,
			TVertexScalar i_left_normal_x, TVertexScalar i_left_normal_y,
			int i_depth,

			const CTsunamiElementData *i_elementData,
			CTsunamiEdgeData *o_edge_comm_data
	)
	{
		*o_edge_comm_data = i_elementData->dofs;
		projectToNormalSpaceAndInvert(o_edge_comm_data->qx, o_edge_comm_data->qy, i_right_normal_x, i_right_normal_y);

#if COMPILE_WITH_DEBUG_TSUNAMI_ELEMENTDATA_VALIDATION
		i_elementData->validation.setupRightEdgeCommData(&(o_edge_comm_data->validation));
		o_edge_comm_data->validation.setupNormal(i_right_normal_x, i_right_normal_y);
#endif
	}



	/**
	 * \brief CALLED by TRAVERSATOR: Store data send via left edge
	 *
	 * this method is responsible for storing the edge communication data
	 * across the left edge.
	 *
	 * depending on the implemented elementData operations, the edge communication
	 * data has to be rotated to the normal space.
	 */
	inline void storeLeftEdgeCommData(
			TVertexScalar i_hyp_normal_x, TVertexScalar i_hyp_normal_y,
			TVertexScalar i_right_normal_x, TVertexScalar i_right_normal_y,
			TVertexScalar i_left_normal_x, TVertexScalar i_left_normal_y,
			int i_depth,

			const CTsunamiElementData *i_elementData,
			CTsunamiEdgeData *o_edge_comm_data
	)
	{
		*o_edge_comm_data = i_elementData->dofs;
		projectToNormalSpaceAndInvert(o_edge_comm_data->qx, o_edge_comm_data->qy, i_left_normal_x, i_left_normal_y);

#if COMPILE_WITH_DEBUG_TSUNAMI_ELEMENTDATA_VALIDATION
		i_elementData->validation.setupLeftEdgeCommData(&(o_edge_comm_data->validation));
		o_edge_comm_data->validation.setupNormal(i_left_normal_x, i_left_normal_y);
#endif
	}




	/**
	 * \brief This method which is executed when all fluxes are computed
	 *
	 * this method is executed whenever all fluxes are computed. then
	 * all data is available to update the elementData within one timestep.
	 */
	inline void computeElementDataWithFlux(
			TVertexScalar i_hyp_normal_x, TVertexScalar i_hyp_normal_y,		///< normals for hypotenuse edge
			TVertexScalar i_right_normal_x, TVertexScalar i_right_normal_y,	///< normals for right edge
			TVertexScalar i_left_normal_x, TVertexScalar i_left_normal_y,	///< normals for left edge (necessary for back-rotation of element)
			int i_depth,

			CTsunamiEdgeData &i_hyp_edge_flux,		///< incoming flux from hypotenuse
			CTsunamiEdgeData &i_right_edge_flux,	///< incoming flux from right edge
			CTsunamiEdgeData &i_left_edge_flux,		///< incoming flux from left edge

			CTsunamiElementData *io_elementData
	)
	{
		T cat_length = getUnitCathetusLengthForDepth(i_depth)*config.square_side_length;
		T hyp_length = getUnitHypotenuseLengthForDepth(i_depth)*config.square_side_length;

#if 1
//		TTsunamiDataScalar damping = CMath::max((T)1.0-(T)2.0*config.delta_timestep/cathetus_length, (T)0.99);
		TTsunamiDataScalar damping = (T)1.0-(config.delta_timestep/cat_length);

		io_elementData->dofs.qx *= damping;
		io_elementData->dofs.qy *= damping;
#endif

		TTsunamiDataScalar tmp = (TTsunamiDataScalar)config.delta_timestep/((TTsunamiDataScalar)0.5*cat_length*cat_length);
		io_elementData->dofs.h -=  tmp * (hyp_length*i_hyp_edge_flux.h + cat_length*(i_right_edge_flux.h + i_left_edge_flux.h));
		io_elementData->dofs.qx -= tmp * (hyp_length*i_hyp_edge_flux.qx + cat_length*(i_right_edge_flux.qx + i_left_edge_flux.qx));
		io_elementData->dofs.qy -= tmp * (hyp_length*i_hyp_edge_flux.qy + cat_length*(i_right_edge_flux.qy + i_left_edge_flux.qy));
	}



	/**
	 * \brief CALLED by TRAVERSATOR: All outer edge data is available and this method
	 * is executed giving, among others, the outer edge data as parameters.
	 */
	inline void elementAction_EEE(
			TVertexScalar i_hyp_normal_x, TVertexScalar i_hyp_normal_y,
			TVertexScalar i_right_normal_x, TVertexScalar i_right_normal_y,
			TVertexScalar i_left_normal_x, TVertexScalar i_left_normal_y,

			int i_depth,								///< depth of recursion

			CTsunamiElementData *io_elementData,		///< pointer to element data

			CTsunamiEdgeData *i_hyp_edgeData_outer,		///< DOF on hyp edge from adjacent triangle
			CTsunamiEdgeData *i_right_edgeData_outer,	///< DOF on right edge from adjacent triangle
			CTsunamiEdgeData *i_left_edgeData_outer		///< DOF on left edge from adjacent triangle
	)
	{
		assert(config.delta_timestep > 0);

#if COMPILE_WITH_DEBUG_TSUNAMI_ELEMENTDATA_VALIDATION
		i_hyp_edgeData_outer->validation.testInverseNormal(i_hyp_normal_x, i_hyp_normal_y);
		i_right_edgeData_outer->validation.testInverseNormal(i_right_normal_x, i_right_normal_y);
		i_left_edgeData_outer->validation.testInverseNormal(i_left_normal_x, i_left_normal_y);

		/*
		 * disable midpoint check for periodic boundaries - simply doesn't make sense :-)
		 */
		io_elementData->validation.testEdgeMidpoints(
				i_hyp_edgeData_outer->validation,
				i_right_edgeData_outer->validation,
				i_left_edgeData_outer->validation
			);
#endif

		if (CMath::isNan(i_hyp_edgeData_outer->h))
		{
			instabilityDetected = true;
			if (!instabilityDetected)
				std::cerr << "INSTABILITY DETECTED!!!" << std::endl;
			return;
		}


		/*
		 * COMPUTE FLUX
		 */
		CTsunamiEdgeData hyp_edge_flux_inner, left_edge_flux_inner, right_edge_flux_inner;
		CTsunamiEdgeData hyp_edge_flux_outer, left_edge_flux_outer, right_edge_flux_outer;
		T hyp_max_wave_speed, left_max_wave_speed, right_max_wave_speed;

		// HYPOTENUSE EDGE
		CTsunamiEdgeData hyp_edgeData_inner = io_elementData->dofs;
		projectToNormalSpace(hyp_edgeData_inner.qx, hyp_edgeData_inner.qy, i_hyp_normal_x, i_hyp_normal_y);

		fluxSolver.computeNetUpdates(
				hyp_edgeData_inner,		*i_hyp_edgeData_outer,
				hyp_edge_flux_inner,	hyp_edge_flux_outer,
				hyp_max_wave_speed,
				config.gravitational_constant
			);

		backprojectFromNormalSpace(hyp_edge_flux_inner.qx, hyp_edge_flux_inner.qy, i_hyp_normal_x, i_hyp_normal_y);


		// RIGHT EDGE
		CTsunamiEdgeData right_edgeData_inner = io_elementData->dofs;
		projectToNormalSpace(right_edgeData_inner.qx, right_edgeData_inner.qy, i_right_normal_x, i_right_normal_y);

		fluxSolver.computeNetUpdates(
				right_edgeData_inner,	*i_right_edgeData_outer,
				right_edge_flux_inner,	right_edge_flux_outer,
				right_max_wave_speed,
				config.gravitational_constant
			);

		backprojectFromNormalSpace(right_edge_flux_inner.qx, right_edge_flux_inner.qy, i_right_normal_x, i_right_normal_y);


		// LEFT EDGE
		CTsunamiEdgeData left_edgeData_inner = io_elementData->dofs;
		projectToNormalSpace(left_edgeData_inner.qx, left_edgeData_inner.qy, i_left_normal_x, i_left_normal_y);

		fluxSolver.computeNetUpdates(
				left_edgeData_inner,	*i_left_edgeData_outer,
				left_edge_flux_inner,	left_edge_flux_outer,
				left_max_wave_speed,
				config.gravitational_constant
			);

		backprojectFromNormalSpace(left_edge_flux_inner.qx, left_edge_flux_inner.qy, i_left_normal_x, i_left_normal_y);

		computeElementDataWithFlux(
				i_hyp_normal_x, i_hyp_normal_y,
				i_right_normal_x, i_right_normal_y,
				i_left_normal_x, i_left_normal_y,
				i_depth,

				hyp_edge_flux_inner,
				right_edge_flux_inner,
				left_edge_flux_inner,

				io_elementData
			);
	}



	/**
	 * \brief CALLED by TRAVERSATOR: Edge types - (hyp,right,left) == (border,comm,comm)
	 */
	inline void elementAction_BEE(
			TVertexScalar hyp_normal_x, TVertexScalar hyp_normal_y,
			TVertexScalar right_normal_x, TVertexScalar right_normal_y,
			TVertexScalar left_normal_x, TVertexScalar left_normal_y,
			int depth,
			CTsunamiElementData *element,
			CTsunamiEdgeData *right_edge,
			CTsunamiEdgeData *left_edge
	);

	/**
	 * \brief CALLED by TRAVERSATOR: Edge types - (hyp,right,left) == (border,border,border)
	 */
	inline void elementAction_BBB(
			TVertexScalar hyp_normal_x, TVertexScalar hyp_normal_y,
			TVertexScalar right_normal_x, TVertexScalar right_normal_y,
			TVertexScalar left_normal_x, TVertexScalar left_normal_y,
			int depth,
			CTsunamiElementData *element
	);


	/**
	 * \brief CALLED by TRAVERSATOR: Edge types - (hyp,right,left) == (comm,comm,border)
	 */
	inline void elementAction_EEB(
			TVertexScalar hyp_normal_x, TVertexScalar hyp_normal_y,
			TVertexScalar right_normal_x, TVertexScalar right_normal_y,
			TVertexScalar left_normal_x, TVertexScalar left_normal_y,
			int depth,
			CTsunamiElementData *element,
			CTsunamiEdgeData *hyp_edge,
			CTsunamiEdgeData *right_edge
	);

	/**
	 * \brief CALLED by TRAVERSATOR: Edge types - (hyp,right,left) == (comm,border,comm)
	 */
	inline void elementAction_EBE(
			TVertexScalar hyp_normal_x, TVertexScalar hyp_normal_y,
			TVertexScalar right_normal_x, TVertexScalar right_normal_y,
			TVertexScalar left_normal_x, TVertexScalar left_normal_y,
			int depth,
			CTsunamiElementData *element,
			CTsunamiEdgeData *hyp_edge,
			CTsunamiEdgeData *left_edge
	);

	/**
	 * \brief CALLED by TRAVERSATOR: Edge types - (hyp,right,left) == (comm,border,border)
	 */
	inline void elementAction_EBB(
			TVertexScalar hyp_normal_x, TVertexScalar hyp_normal_y,
			TVertexScalar right_normal_x, TVertexScalar right_normal_y,
			TVertexScalar left_normal_x, TVertexScalar left_normal_y,
			int depth,
			CTsunamiElementData *element,
			CTsunamiEdgeData *hyp_edge
	);

	/**
	 * \brief CALLED by TRAVERSATOR: Edge types - (hyp,right,left) == (border,border,comm)
	 */
	inline void elementAction_BBE(
			TVertexScalar hyp_normal_x, TVertexScalar hyp_normal_y,
			TVertexScalar right_normal_x, TVertexScalar right_normal_y,
			TVertexScalar left_normal_x, TVertexScalar left_normal_y,
			int depth,
			CTsunamiElementData *element,
			CTsunamiEdgeData *left_edge
	);

	/**
	 * \brief CALLED by TRAVERSATOR: Edge types - (hyp,right,left) == (border,comm,border)
	 */
	inline void elementAction_BEB(
			TVertexScalar hyp_normal_x, TVertexScalar hyp_normal_y,
			TVertexScalar right_normal_x, TVertexScalar right_normal_y,
			TVertexScalar left_normal_x, TVertexScalar left_normal_y,
			int depth,
			CTsunamiElementData *element,
			CTsunamiEdgeData *right_edge
	);
};

#include "CEdgeComm_Tsunami_0thOrder_BoundaryConditions.hpp"

}
}



#endif
